from ktrader.candle_trader.component.rule import Rule
from ktrader import utils


class Record():
    def __init__(self, candle, rule: Rule, index=0):
        self.candle = candle
        self.rule = rule
        self.money = rule.money
        self.index = index
        self.currentOrderDatas = []
        self.historyOrderDatas = []
        self.minutes = [utils.date.ts_to_minute(ts) for ts in candle[:, 0]]
        self.last_sellLine = {
            'buy': None,
            'sell': None,
        }

    def update(self):
        if self.minutes[self.index] == 0:
            self.day_index = self.index
            self.day_open = self.candle[self.index, 1]
            self.day_high = self.candle[self.index, 2]
            self.day_low = self.candle[self.index, 3]

        else:
            # try:
            #     self.day_high = self.candle[self.day_index:self.index, 2].max()
            #     self.day_low = self.candle[self.day_index:self.index, 3].min()
            # except:
            #     self.day_high = None
            #     self.day_low = None
            if hasattr(self, 'day_index'):
                self.day_high = self.candle[self.day_index:self.index, 2].max()
                self.day_low = self.candle[self.day_index:self.index, 3].min()
            else:
                self.day_high = None
                self.day_low = None

        if self.index > 0:
            self.last_open = self.candle[self.index - 1, 1]
            self.last_high = self.candle[self.index - 1, 2]
            self.last_low = self.candle[self.index - 1, 3]
            self.last_close = self.candle[self.index - 1, 4]
        else:
            self.last_open = None
            self.last_high = None
            self.last_low = None
            self.last_close = None

        self.open = self.candle[self.index, 1]
        self.high = self.candle[self.index, 2]
        self.low = self.candle[self.index, 3]
        self.close = self.candle[self.index, 4]

        for currentOrderData in self.currentOrderDatas:
            currentOrderData['processMax'] = max(self.high, currentOrderData['processMax'])
            currentOrderData['processMin'] = min(self.low, currentOrderData['processMin'])
            currentOrderData['processMaxV'] = round(
                (currentOrderData['processMax'] - currentOrderData['buyLine']) / currentOrderData['buyLine'], 5)
            currentOrderData['processMinV'] = round(
                (currentOrderData['processMin'] - currentOrderData['buyLine']) / currentOrderData['buyLine'], 5)

    def get_minute(self):
        return self.minutes[self.index]



from ktrader.candle_trader.component.record import Record


class Period():

    def __init__(self, record):
        self.__record: Record = record

    def isPeriodAllowd(self, index, periods):
        if not periods or not periods[0]:
            return False
        for [minute_start, minute_end] in periods:
            if self.__record.minutes[index] >= minute_start and self.__record.minutes[index] <= minute_end:
                return True
        return False

    def isPeriodBuy(self, index=None):
        return self.isPeriodAllowd(
            index=index or self.__record.index,
            periods=self.__record.rule.BUY_PERIODS,
        )

    def isPeriodSell(self, index=None):
        return self.isPeriodAllowd(
            index=index or self.__record.index,
            periods=self.__record.rule.SELL_PERIODS,
        )

    def isPeriodThrow(self, index=None):
        return self.isPeriodAllowd(
            index=index or self.__record.index,
            periods=self.__record.rule.THROW_PERIODS,
        )



from ktrader import utils


class Rule():
    # 【1.购买、卖出、抛出】---------------------
    # TP = 0.005
    # SL = 0.04  # 正负号问题

    SIDE_BUY_TP =  0.005
    SIDE_BUY_SL = 0.01

    SIDE_SELL_TP = 0.015
    SIDE_SELL_SL = 0.035


    # 运行购买的时间
    BUY_PERIODS = [
        # ['00:00:00', '23:59:00']
    ]
    # 运行卖出的时间
    SELL_PERIODS = [
        # ['00:00:00', '23:59:59']
    ]
    # 运行抛出的时间
    THROW_PERIODS = [
        # ['23:59:00', '23:59:00'],
    ]

    # 手续费率
    COMMISSION_RATE = 0.0003
    # 本金
    money = 100
    # 杠杆
    LEVEL = 10

    def __init__(self):
        BUY_PERIODS_MINUTE = []
        if not self.BUY_PERIODS or not self.BUY_PERIODS[0]:
            self.BUY_PERIODS = [['00:00:00', '23:59:59']]
        for buy_period in self.BUY_PERIODS:
            BUY_PERIODS_MINUTE.append(
                [
                    utils.date.time_to_minute(buy_period[0]),
                    utils.date.time_to_minute(buy_period[1]),
                ],
            )
        self.BUY_PERIODS = BUY_PERIODS_MINUTE

        SELL_PERIODS_MINUTE = []
        if not self.SELL_PERIODS or not self.SELL_PERIODS[0]:
            self.SELL_PERIODS = [['00:00:00', '23:59:59']]

        for sell_period in self.SELL_PERIODS:
            SELL_PERIODS_MINUTE.append(
                [
                    utils.date.time_to_minute(sell_period[0]),
                    utils.date.time_to_minute(sell_period[1]),
                ],
            )
        self.SELL_PERIODS = SELL_PERIODS_MINUTE

        THROW_PERIODS_MINUTE = []
        if self.THROW_PERIODS and self.THROW_PERIODS[0]:
            for throw_period in self.THROW_PERIODS:
                THROW_PERIODS_MINUTE.append(
                    [
                        utils.date.time_to_minute(throw_period[0]),
                        utils.date.time_to_minute(throw_period[1]),
                    ],
                )
        self.THROW_PERIODS = THROW_PERIODS_MINUTE



import datetime


def ts_to_datetime(ts, format='%Y-%m-%d %H:%M:%S'):
    orderDatetime = datetime.datetime.fromtimestamp(int(ts / 1000)).strftime(format)
    return orderDatetime


def datetime_to_ts(date):
    return datetime.datetime.strptime(date, '%Y-%m-%d %H:%M:%S').timestamp() * 1000


def ts_to_minute(ts):
    date = datetime.datetime.fromtimestamp(int(ts / 1000))
    return date.hour * 60 + date.minute


def time_to_minute(time):
    date = datetime.datetime.strptime(time, '%H:%M:%S')
    return date.hour * 60 + date.minute


# 计算两个字符串日期对象的时间间隔
def derta_datetime_minute(date1, date2):
    timestamp1 = datetime.datetime.strptime(date1, '%Y-%m-%d %H:%M:%S').timestamp()
    timestamp2 = datetime.datetime.strptime(date2, '%Y-%m-%d %H:%M:%S').timestamp()
    return abs(int((timestamp1 - timestamp2) / 60))




from ktrader.candle_trader.component import period, order, rule, trade, record, analysis
from ktrader import utils

# from importlib import reload
# reload(period)
# reload(order)
# reload(rule)
# reload(trade)
# reload(utils)
# reload(record)
# reload(analysis)



class CandlerTrader():
    def __init__(self, candle, rule):
        self.rule = rule
        self.record = record.Record(candle=candle, rule=rule, index=0)
        self.period = period.Period(record=self.record)
        self.order = order.Order(self.record)
        self.trade = trade.Trade(record=self.record, period=self.period, order=self.order)
        self.analysis = analysis.Analysis(record=self.record, order=self.order)
        self.setup()

    def setup(self):
        # to override
        pass

    def buy(self, *args, **kwargs):
        # to override
        pass

    def sell(self, *args, **kwargs):
        # to override
        pass

    def run(self, start_date=None, end_date=None, clear=True):
        start_index = utils.candle.get_index_by_date(
            candle=self.record.candle,
            date=start_date,
            default=0
        )
        end_index = utils.candle.get_index_by_date(
            candle=self.record.candle,
            date=end_date,
            default=self.record.candle.shape[0]
        )

        for index in range(start_index, end_index - 1):
            self.index = self.record.index = index
            self.record.update()

            self.order.to_historyOrderDatas(self.trade.periodThrow())
            tpSellOrders = [order for order in self.trade.tpSell()]
            self.order.to_historyOrderDatas(tpSellOrders)

            slThrowOrders = [order for order in self.trade.slThrow()]
            self.order.to_historyOrderDatas(slThrowOrders)
            if self.period.isPeriodSell() and not slThrowOrders:
                self.order.to_historyOrderDatas(self.sell())
            if self.period.isPeriodBuy() and not tpSellOrders and not slThrowOrders:
                self.order.to_currentOrderDatas(self.buy())

        if clear:
            self.order.to_historyOrderDatas(self.trade.marketThrow(self.record.currentOrderDatas))

    # 调用
    def start_orderData(self, posSide, buyLine, buyMoney, orderType=None, tpRate=None, slRate=None, firm=True,
                        **kwargs):
        return self.order.start_orderData(posSide=posSide, buyLine=buyLine, buyMoney=buyMoney, orderType=orderType,
                                          tpRate=tpRate, slRate=slRate, firm=firm, **kwargs)

    def close_orderData(self, orderData, sellLine, endType=None, index=None, firm=True, **kwargs):
        return self.order.close_orderData(orderData=orderData, sellLine=sellLine, endType=endType, index=index,
                                          firm=firm, **kwargs)

    # attribute ---------------------------------
    @property
    def open(self):
        return self.record.open

    @property
    def high(self):
        return self.record.high

    @property
    def low(self):
        return self.record.low

    @property
    def close(self):
        return self.record.close

    @property
    def day_index(self):
        if hasattr(self.record, 'day_index'):
            return self.record.day_index
        else:
            return None

    @property
    def day_open(self):
        return self.record.day_open

    @property
    def day_high(self):
        return self.record.day_high

    @property
    def day_low(self):
        return self.record.day_low

    '''
    todo 当前时间，当前分钟
    '''



from ktrader.candle_trader.component.record import Record
from ktrader.utils.price import round_px
from ktrader import utils
from copy import deepcopy
import warnings


class Order():
    def __init__(self, record):
        self.__record: Record = record

    @staticmethod
    def get_empty_orderData():
        '''
        init需要更新的参数：
            orderType       订单类型 str
            posSide         方向 long做多，short做空
            buyIndex        购买时候的index int
            orderDatetime   订单开始的时间，%Y-m-%d %H:%M:%S str
            buyLine         购买的价格 float
            tpLine          止盈的价格 float
            slLine          止损的价格 float
            volCcy24        24小时的USDT成交量 float
            buyMoney        买入的金额 float
        close需要更新的参数：
            endType         订单终止的方式
            sellIndex       卖出时候的index int
            endDatetime     订单终止的时间，%Y-%m-%d %H:%M:%S str
            sellLine        卖出的价格 float
            holdMinute      从订单开始到订单结束的时间，单位分钟 int
            sellMoney       卖出的金额（包含抛出） float
            diff            利润率 float
            profit          利润 float
        process需要更新的参数：
            processMax     订单开始到订单终止的最大价格
            processMaxV    订单开始到订单终止的最大涨幅 float
            processMin     订单开始到订单终止的最小价格
            processMinV    订单开始到订单终止的最大跌幅 float
        '''
        orderData = {
            # 订单类型 str
            'orderType': None,
            # 方向：long做多，short做空
            'posSide': None,
            # 购买时候的index int
            'buyIndex': None,
            # 订单开始的时间，%Y-m-%d %H:%M:%S, str
            'orderDatetime': None,
            # 购买的价格 float
            'buyLine': None,
            # 止盈的价格 float
            'tpLine': None,
            # 止损的价格 float
            'slLine': None,
            # 24小时的USDT成交量 float
            'volCcy24': None,
            # 买入的金额 float
            'buyMoney': None,
            # 订单终止的方式 str
            'endType': None,
            # 卖出时候的index int
            'sellIndex': None,
            # 订单终止的时间，%Y-%m-%d %H:%M:%S str
            'endDatetime': None,  # str
            # 卖出的价格 float
            'sellLine': None,
            # 从订单开始到订单结束的时间，单位分钟 int
            'holdMinute': None,
            # 卖出的金额（包含抛出） float
            'sellMoney': None,
            # 利润率 float
            'diff': None,
            # 手续费
            'commission': None,
            # 利润 float
            'profit': None,
            #  订单开始到订单终止的最大价格
            'processMax': None,
            #  订单开始到订单终止的最大涨幅 float
            'processMaxV': None,
            #  订单开始到订单终止的最小价格
            'processMin': None,
            # 订单开始到订单终止的最大跌幅 float
            'processMinV': None,
        }
        return deepcopy(orderData)

    def to_historyOrderDatas(self, orderDatas):
        if orderDatas:
            for orderData in orderDatas:
                try:
                    self.__record.currentOrderDatas.remove(orderData)
                    self.__record.historyOrderDatas.append(orderData)
                    self.__record.money += orderData['sellMoney']
                    # if orderData['posSide'] == 'long':
                    #     self.__record.last_sellLine['buy'] = orderData['sellLine']
                    # elif orderData['posSide'] == 'short':
                    #     self.__record.last_sellLine['sell'] = orderData['sellLine']
                except:
                    pass

    def to_currentOrderDatas(self, orderDatas):
        if orderDatas:
            for orderData in orderDatas:
                self.__record.currentOrderDatas.append(orderData)
                self.__record.money -= orderData['buyMoney']

    def start_orderData(self, posSide, buyLine, buyMoney, orderType=None, tpRate=None, slRate=None, firm=True,
                        **kwargs):
        orderData = self.get_empty_orderData()
        buyLine = round_px(buyLine)
        # 验证是否能以buyLine进行购买
        if buyLine >= self.__record.low and buyLine <= self.__record.high:
            orderData['buyLine'] = buyLine
        # 如果不能成功购买
        else:
            # firm等于True、则会收盘的时候购买
            if firm == True:
                buyLine = self.__record.close
                orderData['buyLine'] = buyLine
                # todo 报告警告
            # 报告异常 todo + ts
            else:
                raise Exception(
                    'buyLine:{buyLine} not between low:{low}~high{high}'.format(
                        buyLine=buyLine,
                        low=self.__record.low,
                        high=self.__record.high,
                    )
                )
        orderData['orderType'] = orderType
        orderData['orderDatetime'] = utils.date.ts_to_datetime(self.__record.candle[self.__record.index, 0])
        orderData['orderTs'] = self.__record.candle[self.__record.index, 0]
        orderData['buyIndex'] = self.__record.index
        if posSide == 'long':
            if tpRate:
                tpLine = buyLine * (1 + abs(tpRate))
                slLine = buyLine * (1 - abs(slRate))
            else:
                tpLine = (abs(self.__record.rule.SIDE_BUY_TP) + 1) * buyLine
                slLine = (-abs(self.__record.rule.SIDE_BUY_SL) + 1) * buyLine
        elif posSide == 'short':
            if tpRate:
                tpLine = buyLine * (1 - abs(tpRate))
                slLine = buyLine * (1 + abs(slRate))
            else:
                tpLine = (-abs(self.__record.rule.SIDE_SELL_TP) + 1) * buyLine
                slLine = (abs(self.__record.rule.SIDE_SELL_SL) + 1) * buyLine
        else:
            raise Exception('error posSide')  # todo
        orderData['tpLine'] = round_px(tpLine)
        orderData['slLine'] = round_px(slLine)
        orderData['buyMoney'] = buyMoney
        orderData['posSide'] = posSide
        orderData['processMax'] = buyLine
        orderData['processMin'] = buyLine
        orderData['processMaxV'] = 0
        orderData['processMinV'] = 0
        if kwargs:
            for key in kwargs:
                orderData[key] = kwargs[key]
        return orderData

    def close_orderData(self, orderData, sellLine, endType=None, index=None, firm=True, **kwargs):
        if not index:
            index = self.__record.index
        sellLine = orderData['sellLine'] = round_px(sellLine)
        # 如果卖出价格不在low~high之间
        if not (sellLine >= self.__record.low and sellLine <= self.__record.high):
            if firm == True:
                sellLine = self.__record.close
                orderData['sellLine'] = sellLine
                # todo + warning
            else:
                # todo + ts
                raise Exception(
                    'sellLine:{sellLine} not between low:{low}~high{high}'.format(
                        sellLine=sellLine,
                        low=self.__record.low,
                        high=self.__record.high,
                    )
                )

        orderData['sellIndex'] = self.__record.index
        orderData['endType'] = endType
        orderData['endDatetime'] = utils.date.ts_to_datetime(self.__record.candle[index, 0])
        orderData['endTs'] = self.__record.candle[self.__record.index, 0]
        orderData['sellLine'] = sellLine
        orderData['holdMinute'] = utils.date.derta_datetime_minute(
            orderData['orderDatetime'], orderData['endDatetime']
        )

        s = orderData['sellLine']
        b = orderData['buyLine']
        r = self.__record.rule.COMMISSION_RATE
        l = self.__record.rule.LEVEL
        if orderData['posSide'] == 'long':
            orderData['commission'] = (l * r + (1 - l * r) * (l * (s - b) + b) / b * l * r) * orderData['buyMoney']
            orderData['diff'] = round((1 - l * r) ** 2 * (l * (s - b) + b) / b - 1, 5)

        elif orderData['posSide'] == 'short':
            orderData['commission'] = l * r + (1 - l * r) * (b + (b - s) * l) / b * l * r * orderData['buyMoney']
            orderData['diff'] = round((1 - l * r) ** 2 * (b + (b - s) * l) / b - 1, 5)
        else:
            raise Exception(orderData['posSide'])
        orderData['sellMoney'] = round(orderData['buyMoney'] * (1 + orderData['diff']), 5)
        orderData['profit'] = round(orderData['diff'] * orderData['buyMoney'], 5)
        # orderData['open'] = self.__record.open
        # orderData['close'] = self.__record.close
        # orderData['high'] = self.__record.high
        # orderData['low'] = self.__record.low

        # 更新过程中的最高与最低价格
        # orderData['processMax'] = max(
        #     self.__record.candle[orderData['buyIndex']:orderData['sellIndex'], 2]
        # )
        # orderData['processMin'] = min(
        #     self.__record.candle[orderData['buyIndex']:orderData['sellIndex'], 3]
        # )
        # orderData['processMaxV'] = round((orderData['processMax'] - orderData['buyLine']) / orderData['buyLine'], 4)
        # orderData['processMinV'] = round((orderData['processMin'] - orderData['buyLine']) / orderData['buyLine'], 4)
        if kwargs:
            for key in kwargs:
                orderData[key] = kwargs[key]
        return orderData



from ktrader.candle_trader.component.record import Record
from ktrader.candle_trader.component.order import Order
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from copy import deepcopy
import datetime


class Analysis():
    def __init__(self, record: Record, order: Order):
        self.__record: Record = record
        self.__order: Order = order

    def get_df(self, **kwargs):
        if not self.__record.historyOrderDatas:
            return None
        default_columns = [
            'orderType',
            'posSide',
            'endType',
            'orderDatetime',
            'endDatetime',
            'buyLine',
            'sellLine',
            'tpLine',
            'slLine',
            'buyMoney',
            'sellMoney',
            'commission',
            'diff',
            'profit',
            'processMaxV',
            'processMinV'
        ]
        df_columns = deepcopy(default_columns)

        for column in kwargs.keys():
            value = kwargs[column]
            if value == True and column not in df_columns:
                df_columns.append(column)
                continue
            if value == False and column in df_columns:
                df_columns.remove(column)
                continue

        # 用户在buy或者sell中添加的字段
        add_columns = list(set([column for column in self.__record.historyOrderDatas[0]]) - set(
            [column for column in self.__order.get_empty_orderData().keys()]))
        for column in add_columns:
            if column not in df_columns:
                df_columns.append(column)
        df = pd.DataFrame(self.__record.historyOrderDatas)[df_columns]
        return df

    def get_global_df(self):
        column_map = {}
        for column in self.__order.get_empty_orderData():
            column_map[column] = True
        self.df = self.get_df(**column_map)
        self.df['day'] = self.df['endDatetime'].apply(
            lambda d: datetime.datetime.strptime(d, '%Y-%m-%d %H:%M:%S').strftime('%Y-%m-%d')
        )

        self.df['month'] = self.df['endDatetime'].apply(
            lambda d: datetime.datetime.strptime(d, '%Y-%m-%d %H:%M:%S').strftime('%Y-%m')
        )
        return self.df

    def df_profit(self, by='day'):
        if not hasattr(self, 'df'):
            self.get_global_df()

        return self.df.groupby(by)['profit', 'diff'].sum().reset_index().sort_values(by=by)

        # return df.groupby('month').aggregate(
        #     {'diff': ['sum', 'mean'],
        #      'profit': ['sum', 'count']}
        # )

    def plot_profit(self, start_datetime=None, end_datetime=None, by='day', cumsum=True,figsize=(20, 8)):
        '''
        :param start_datetime:
        :param end_datetime:
        :param by: index|day
        :return:
        '''
        if not hasattr(self, 'df'):
            self.get_global_df()
        if not start_datetime:
            start_datetime = '0000-00-00 00:00:00'
        if not end_datetime:
            end_datetime = '9999-99-99 00:00:00'
        df: pd.DataFrame = self.df.query('orderDatetime >= @start_datetime and endDatetime <= @end_datetime')

        plt.rcParams['font.sans-serif'] = ['Arial Unicode MS']
        plt.rcParams['font.size'] = 15
        plt.figure(figsize=figsize)
        linewidth = max(figsize[0] / 20 * 1, 1)
        pointsize = max(figsize[0] / 20 * 100, 100)
        ticksfontsize = 10 * figsize[0] / 20
        labelfontsize = 15 * figsize[0] / 20
        titlefontsize = 15 * figsize[0] / 20

        if by == 'index':
            x = df['endTs'].values
            profits = df['profit'].values
            if cumsum:
                profits = np.cumsum(profits)
            plt.plot(x, profits,linewidth=linewidth)
            xticks = []
            xlabels = []
            ts_min = self.__record.candle[:, 0].min()
            ts_max = self.__record.candle[:, 0].max()
            start_date = datetime.datetime.fromtimestamp(int(ts_min / 1000))
            end_date = datetime.datetime.fromtimestamp(int(ts_max / 1000))
            derta_day = (end_date - start_date).days
            for day in range(derta_day + 3):
                date_str = (start_date + datetime.timedelta(days=day)).strftime('%Y-%m-%d')
                date_ts = datetime.datetime.strptime(date_str, '%Y-%m-%d').timestamp() * 1000
                xticks.append(date_ts)
                xlabels.append(date_str)
            plt.xticks(ticks=xticks, labels=xlabels, rotation=90, fontsize=ticksfontsize)
        elif by == 'day':
            series_day = df.groupby('day')['profit'].sum()
            x = series_day.index
            profits = series_day.values
            if cumsum:
                profits = np.cumsum(profits)
            plt.plot(x, profits,linewidth=linewidth)
            plt.xticks(ticks=x, labels=x, rotation=90, fontsize=ticksfontsize)

        else:
            raise Exception('by not in [index,day]')
        plt.xlabel('日期', fontsize=labelfontsize)
        plt.ylabel('利润', fontsize=labelfontsize)
        plt.title('交易利润变化', fontsize=titlefontsize)
        plt.plot(x, np.zeros_like(x), linewidth=linewidth * 0.5)
        plt.grid(linewidth = linewidth * 0.5)
        plt.show()


if __name__ == '__main__':
    ret = np.cumsum([1, 2, 3, 10])
    print(ret)



def round_px(px):
    if px == 0:
        return px
    px_0_0001 = 0.0001 * px
    if px_0_0001 < 1:
        ndigits = len(str(int(1 / px_0_0001)))
        px_round = round(px, ndigits)
    else:
        px_round = round(px, 1)
    if abs((px_round - px) / px) >= 0.0001:
        raise Exception(
            'px_round={px_round},px={px}'.format(
                px_round=px_round,
                px=px
            )
        )
    else:
        return px_round



from ktrader.candle_trader.component.record import Record
from ktrader.candle_trader.component.period import Period
from ktrader.candle_trader.component.order import Order


class Trade():
    def __init__(self, record, period, order):
        self.__record: Record = record
        self.__period: Period = period
        self.__orderData: Order = order

    # 止盈卖出
    def tpSell(self):
        index = self.__record.index
        for orderData in self.__record.currentOrderDatas:
            posSide = orderData['posSide']
            # 做多与做多
            if (posSide == 'long' and self.__record.high >= orderData['tpLine']) or \
                    (posSide == 'short' and self.__record.low <= orderData['tpLine']):
                sellOrderData = self.__orderData.close_orderData(
                    orderData=orderData, endType='tp', sellLine=orderData['tpLine'], index=index
                )
                yield sellOrderData

    # 止损抛出
    def slThrow(self):
        # index = self.__record.index
        # for orderData in self.__record.currentOrderDatas:
        #     side = orderData['side']
        # # 做多与做多
        #     if (side == 'buy' and self.__record.low <= orderData['slLine']) or \
        #             (side == 'sell' and self.__record.high >= orderData['slLine']):
        #         throwOrderData = self.__orderData.close_orderData(
        #             orderData=orderData, endType='sl', sellLine=orderData['slLine'], index=index
        #         )
        #         yield throwOrderData
        index = self.__record.index
        for orderData in self.__record.currentOrderDatas:
            posSide = orderData['posSide']
            # 做多与做多
            if (posSide == 'long' and self.__record.open <= orderData['slLine']) or \
                    (posSide == 'short' and self.__record.open >= orderData['slLine']):
                throwOrderData = self.__orderData.close_orderData(
                    orderData=orderData, endType='sl', sellLine=self.__record.close, index=index
                )
                yield throwOrderData

    # 按照时间段抛出全部
    def periodThrow(self):
        index = self.__record.index
        if self.__period.isPeriodThrow():
            for orderData in self.__record.currentOrderDatas:
                periodThrowOrderData = self.__orderData.close_orderData(
                    orderData, endType='pt', sellLine=self.__record.open, index=index,
                )
                yield periodThrowOrderData

    def marketThrow(self, orderDatas):
        for orderData in orderDatas:
            marketThrowOrderData = self.__orderData.close_orderData(
                orderData, endType='mt', sellLine=self.__record.open
            )
            yield marketThrowOrderData

    def buy(self, *args, **kwargs):
        pass

    def sell(self, *args, **kwargs):
        pass



from ktrader import candle_trader
from ktrader.candle_trader.component import rule
import numpy as np
import matplotlib.pyplot as plt
import datetime


# import importlib
# importlib.reload(candle_trader)


def isNone(obj):
    return type(obj).__name__ == 'NoneType'


def get_dualThrust_array(candle, n: int, ks: float, kx: float):
    '''
    :param: candle
    :param: 前n分钟
    '''
    dualThrust_list = []
    for i in range(n, candle.shape[0], n):
        # 前bar的数据
        last_k_array = candle[i - n:i]
        hh = last_k_array[:, 2].max()  # 最高价的最高价
        hc = last_k_array[:, 4].max()  # 最高价的收盘价
        lc = last_k_array[:, 4].min()  # 最低价的收盘价
        ll = last_k_array[:, 3].min()  # 最低价的最低价
        range_ = max(hh - lc, hc - ll)
        open = candle[i, 1]  # 开盘价
        range_ks = open + ks * range_
        range_kx = open - kx * range_
        # 还剩下的数据长度
        length = min(n, candle.shape[0] - i)
        for ts in candle[i:i + length, 0]:
            dualThrust_list.append([ts, range_ks, range_kx, hh, hc, lc, ll])
    return np.array(dualThrust_list)


class Rule(rule.Rule):
    LEVEL = 1


class DualThrust(candle_trader.CandlerTrader):
    def __init__(self, candle, n, ks, kx, position=2, money=1, reverse=False, rule=Rule()):
        self.reverse = reverse
        self.position = position
        self.money = money
        self.n = n
        self.ks = ks
        self.kx = kx
        self.dualThrust_array = get_dualThrust_array(
            candle=candle,
            n=self.n,
            ks=self.ks,
            kx=self.kx,
        )
        candle = candle[n:]  # 去掉前n行的candle
        super(DualThrust, self).__init__(candle=candle, rule=rule)

    @property
    def range_ks(self):
        return self.dualThrust_array[self.index, 1]

    @property
    def range_kx(self):
        return self.dualThrust_array[self.index, 2]

    def buy_wapper(self):
        if self.reverse == False:
            # 做多
            if not hasattr(self, 'latest_buy_range_ks') or self.range_ks != self.latest_buy_range_ks:
                if self.high >= self.range_ks:
                    yield self.start_orderData(
                        orderType='dualThrust',
                        posSide='long',
                        buyLine=self.range_ks,
                        buyMoney=self.money,
                        tpRate=1,
                        slRate=1,
                    )
                    self.latest_buy_range_ks = self.range_ks
            # 做空
            if not hasattr(self, 'latest_buy_range_kx') or self.range_kx != self.latest_buy_range_kx:
                if self.low <= self.range_kx:
                    yield self.start_orderData(
                        orderType='dualThrust',
                        posSide='short',
                        buyLine=self.range_kx,
                        buyMoney=self.money,
                        tpRate=1,
                        slRate=1,
                    )
                    self.latest_buy_range_kx = self.range_kx
        elif self.reverse:
            # 做空
            if not hasattr(self, 'latest_buy_range_ks') or self.range_ks != self.latest_buy_range_ks:
                if self.high >= self.range_ks:
                    yield self.start_orderData(
                        orderType='dualThrust',
                        posSide='short',
                        buyLine=self.range_ks,
                        buyMoney=self.money,
                        tpRate=1,
                        slRate=1,
                    )
                    self.latest_buy_range_ks = self.range_ks
            # 做空
            if not hasattr(self, 'latest_buy_range_kx') or self.range_kx != self.latest_buy_range_kx:
                if self.low <= self.range_kx:
                    yield self.start_orderData(
                        orderType='dualThrust',
                        posSide='long',
                        buyLine=self.range_kx,
                        buyMoney=self.money,
                        tpRate=1,
                        slRate=1,
                    )
                    self.latest_buy_range_kx = self.range_kx

    def buy(self):
        orderDatas = [orderData for orderData in self.buy_wapper()]
        for i in range(self.position - len(self.record.currentOrderDatas)):
            if i < len(orderDatas):
                yield orderDatas[i]

    def sell(self):
        if self.reverse == False:
            for orderData in self.record.currentOrderDatas:
                # 多单
                if orderData['posSide'] == 'long' and self.low <= self.range_kx:
                    ts = self.record.candle[self.index, 0]
                    yield self.close_orderData(
                        orderData=orderData,
                        sellLine=self.range_kx,
                    )
                # 空单
                elif orderData['posSide'] == 'short' and self.high >= self.range_ks:
                    yield self.close_orderData(
                        orderData=orderData,
                        sellLine=self.range_ks,
                    )
        else:
            for orderData in self.record.currentOrderDatas:
                # 多单
                if orderData['posSide'] == 'long' and self.high >= self.range_ks:
                    ts = self.record.candle[self.index, 0]
                    yield self.close_orderData(
                        orderData=orderData,
                        sellLine=self.range_kx,
                    )
                # 空单
                elif orderData['posSide'] == 'short' and self.low <= self.range_kx:
                    yield self.close_orderData(
                        orderData=orderData,
                        sellLine=self.range_ks,
                    )

    def plot(self, open=1, range_ks=1, range_kx=1, point=1, figsize=(20, 8)):
        plt.rcParams['font.sans-serif'] = ['Arial Unicode MS']
        plt.rcParams['font.size'] = 15
        plt.figure(figsize=figsize)
        legends = []
        linewidth = max(figsize[0] / 20 * 1, 1)
        pointsize = max(figsize[0] / 20 * 100, 100)
        ticksfontsize = 10 * figsize[0] / 20
        labelfontsize = 15 * figsize[0] / 20
        titlefontsize = 15 * figsize[0] / 20

        ts = self.record.candle[:, 0]
        if open:
            open_line = self.record.candle[:, 1]
            plt.plot(ts, open_line, linewidth=linewidth)
            legends.append('开盘价格')
        if range_ks:
            ks_line = self.dualThrust_array[:, 1]
            plt.plot(ts, ks_line, linewidth=linewidth)
            legends.append('上轨range_ks')
        if range_kx:
            kx_line = self.dualThrust_array[:, 2]
            plt.plot(ts, kx_line, linewidth=linewidth)
            legends.append('下轨range_kx')
        if point:
            long_buy_points = []
            long_sell_points = []
            short_buy_points = []
            short_sell_points = []
            for historyOrderData in self.record.historyOrderDatas:
                posSide = historyOrderData['posSide']
                buyIndex = historyOrderData['buyIndex']
                sellIndex = historyOrderData['sellIndex']
                buyLine = historyOrderData['buyLine']
                sellLine = historyOrderData['sellLine']
                buyTs = self.record.candle[buyIndex, 0]
                sellTs = self.record.candle[sellIndex, 0]
                if posSide == 'long':
                    long_buy_points.append([buyTs, buyLine])
                    long_sell_points.append([sellTs, sellLine])
                elif posSide == 'short':
                    short_buy_points.append([buyTs, buyLine])
                    short_sell_points.append([sellTs, sellLine])

            long_buy_points = np.array(long_buy_points)
            long_sell_points = np.array(long_sell_points)

            short_buy_points = np.array(short_buy_points)
            short_sell_points = np.array(short_sell_points)

            if not isNone(long_buy_points):
                buy_points_ts = long_buy_points[:, 0]
                buy_points_price = long_buy_points[:, 1]
                plt.scatter(buy_points_ts, buy_points_price * 1.001, color='r', s=pointsize, marker='^')
                legends.append('买入开多')

            if not isNone((long_sell_points)):
                sell_points_ts = long_sell_points[:, 0]
                sell_points_price = long_sell_points[:, 1]
                plt.scatter(sell_points_ts, sell_points_price * 0.999, color='k', s=pointsize, marker='^')
                legends.append('卖出平多')

            if not isNone(short_buy_points):
                buy_points_ts = short_buy_points[:, 0]
                buy_points_price = short_buy_points[:, 1]
                plt.scatter(buy_points_ts, buy_points_price * 1.001, color='r', s=pointsize, marker='v')
                legends.append('卖出开空')

            if not isNone((short_sell_points)):
                sell_points_ts = short_sell_points[:, 0]
                sell_points_price = short_sell_points[:, 1]
                plt.scatter(sell_points_ts, sell_points_price * 0.999, color='k', s=pointsize, marker='v')
                legends.append('买入平空')
        xticks = []
        xlabels = []
        ts_min = self.record.candle[:, 0].min()
        ts_max = self.record.candle[:, 0].max()
        start_date = datetime.datetime.fromtimestamp(int(ts_min / 1000))
        end_date = datetime.datetime.fromtimestamp(int(ts_max / 1000))
        derta_day = (end_date - start_date).days
        for day in range(derta_day + 3):
            date_str = (start_date + datetime.timedelta(days=day)).strftime('%Y-%m-%d')
            date_ts = datetime.datetime.strptime(date_str, '%Y-%m-%d').timestamp() * 1000
            xticks.append(date_ts)
            xlabels.append(date_str)
        plt.xticks(ticks=xticks, labels=xlabels, rotation=90, fontsize=10 * figsize[0] / 20)
        plt.xlabel('日期', fontsize=labelfontsize)
        plt.ylabel('价格', fontsize=labelfontsize)
        plt.title('交易过程历史K线图', fontsize=titlefontsize)
        plt.grid(linewidth=linewidth * 0.5)
        plt.legend(legends)

        plt.show()


if __name__ == '__main__':
    from kqt_tool.load import load_array_map_by_date
    from importlib import reload

    candle_map = load_array_map_by_date(start_date='2022-01-01', end_date='2022-01-10', instIds=['BTC-USDT'])
    BTC = candle_map['BTC-USDT']

    from ktrader.candle_trader.component import rule

    reload(rule)

    R = rule.Rule()

    dt = DualThrust(candle=BTC, rule=R, n=60 * 12, ks=0.3, kx=0.3, position=1, reverse=True)

    dt.run()
    # for order in dt.record.historyOrderDatas:
    #     print(order)
    dt.plot()
    dt.analysis.plot_profit(by='day')
    print(dt.analysis.df_profit(by='day'))



import numpy as np
from .date import datetime_to_ts


def get_index_by_date(candle, date, default=0):
    if not date:
        return default
    else:
        ts = datetime_to_ts(date)
        index = np.where(
            candle[:, 0] == ts
        )[0][0]
        return index